/* FreeTDS - Library of routines accessing Sybase and Microsoft databases
 * Copyright (C) 1998-1999  Brian Bruns
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <config.h>
#include "tdsutil.h"
#include "tds.h"
#include "tdsconvert.h"
#include <time.h>
#include <assert.h>
#include <errno.h>

static char  software_version[]   = "$Id: convert.c,v 1.16 2002/06/26 01:44:27 jklowden Exp $";
static void *no_unused_var_warn[] = {software_version,
                                     no_unused_var_warn};

typedef union dbany {
	TDS_TINYINT	ti;
	TDS_SMALLINT	si;
	TDS_INT		i;
	TDS_FLOAT		f;
	TDS_REAL		r;
	TDS_CHAR		*c;
	TDS_MONEY         m;
	TDS_MONEY4	m4;
	TDS_DATETIME	dt;
	TDS_DATETIME4	dt4;
	TDS_NUMERIC	n;
    TDS_UNIQUE  u;
} DBANY; 

struct diglist {
	short int dig;
	short int carried;
	struct diglist *nextptr;
};

typedef unsigned short utf16_t;

static int _tds_pad_string(char *dest, int destlen);
extern char *tds_numeric_to_string(TDS_NUMERIC *numeric, char *s);
extern char *tds_money_to_string(TDS_MONEY *money, char *s);
TDS_INT tds_convert_any(unsigned char *dest, TDS_INT dtype, TDS_INT dlen, DBANY *any);
static int string_to_datetime(char *datestr, int desttype, DBANY *);
static void string_to_numeric(char *instr, DBANY *any);
static int store_hour(char *, char *, struct tds_time *);
static int store_time(char *, struct tds_time * );
static int store_yymmdd_date(char *, struct tds_time *);
static int store_monthname(char *, struct tds_time *);
static int store_numeric_date(char *, struct tds_time *);
static int store_mday(char *, struct tds_time *);
static int store_year(int,  struct tds_time *);
static int is_timeformat(char *);
static int is_numeric(char *);
static int is_alphabetic(char *);
static int is_ampm(char *);
static int is_monthname(char *);
static int is_numeric_dateformat(char *);

extern int g__numeric_bytes_per_prec[];

/* 
this needs to go... 
it won't handle binary or text/image when they are added
it's not thread safe
it works for the moment though til i decide how i really want to handle it
*/
static TDS_CHAR tmp_str[4096];  


int tds_get_conversion_type(int srctype, int colsize)
{
	if (srctype == SYBINTN) {
		if (colsize==8)
			 return SYBINT8;
		else if (colsize==4)
			 return SYBINT4;
		else if (colsize==2)
			 return SYBINT2;
		else if (colsize==1)
			 return SYBINT1;
	} else if (srctype == SYBFLTN) {
		if (colsize==8)
			return SYBFLT8;
		else if (colsize==4)
			return SYBREAL;
	} else if (srctype == SYBDATETIMN) {
		if (colsize==8) 
			return SYBDATETIME;
		else if (colsize==4)
			return SYBDATETIME4;
	} else if (srctype == SYBMONEYN) {
		if (colsize==8) 
			return SYBMONEY;
		else if (colsize==4) 
			return SYBMONEY4;
	}
	return srctype;
}

static TDS_INT 
tds_convert_text(int srctype,TDS_CHAR *src,TDS_UINT srclen,
	int desttype,TDS_CHAR *dest,TDS_UINT destlen)
{
int cplen;

   switch(desttype) {
      case SYBTEXT:
      case SYBCHAR:
	 cplen = srclen > destlen ? destlen : srclen;
         memcpy(dest, src, cplen);
         dest[cplen-1] = '\0';
         return strlen(dest);
      case SYBBINARY:
	 cplen = srclen > destlen ? destlen : srclen;
         memcpy(dest, src, cplen);
         return cplen;
   }
   return 0;
}
static TDS_UINT 
utf16len(const utf16_t* s)
{
      const utf16_t* p = s;
      while (*p++)
              ;
      return p - s;
}

static TDS_INT 
tds_convert_ntext(int srctype,TDS_CHAR *src,TDS_UINT srclen,
      int desttype,TDS_CHAR *dest,TDS_UINT destlen)
{
      /* 
       * XXX Limit of 255 + 1 needs to be fixed by determining what the 
       *     real limit of [N][VAR]CHAR columns is.
       * XXX What about NCHAR?  Don't see a constant for it in tds.h.
       * XXX Case for -1 in switch statement because upper levels don't
       *     have a way to bind to wide-character types.
       */
      TDS_UINT i, cplen, char_limit = 256;
      utf16_t* wsrc  = (utf16_t*)src;
      utf16_t* wdest = (utf16_t*)dest;
      assert(sizeof(utf16_t) == 2);
      switch (desttype) {
              case SYBNVARCHAR:
                      if (destlen > char_limit * sizeof(utf16_t))
                              destlen = char_limit * sizeof(utf16_t);
                      /* Fall through ... */
              case SYBNTEXT:
              case -1:
                      cplen = srclen > destlen ? destlen : srclen;
                      memcpy(dest, src, cplen);
                      if (destlen < srclen + sizeof(utf16_t)) {
                              size_t term_pos  = destlen - sizeof(utf16_t);
                              size_t odd_bytes = term_pos % sizeof(utf16_t);
                              term_pos -= odd_bytes;
                              wdest[term_pos / sizeof(utf16_t)] = 0;
                      }
            else
                              wdest[cplen / sizeof(utf16_t)] = 0;
                      return utf16len(wdest) * 2;
              default:
                      /* Assume caller wants conversion to narrow string. */
                      if (destlen > char_limit && desttype != SYBTEXT)
                              destlen = char_limit;
                      cplen = srclen > destlen ? destlen : srclen;
                      for (i = 0; i < cplen; ++i)
                              dest[i] = (unsigned char)wsrc[i];
                      dest[cplen-1] = 0;
                      return strlen(dest);
      }
      return 0;
}


static TDS_INT 
tds_convert_binary(int srctype,TDS_CHAR *src,TDS_INT srclen,
	int desttype,TDS_CHAR *dest,TDS_INT destlen)
{
int cplen;
int d, s;
TDS_VARBINARY *varbin;

   switch(desttype) {
      case SYBCHAR:
      case SYBVARCHAR:
         /* FIX ME */
	 if (destlen>=0 && destlen<3) {
		return 0;
	 }
	 d=0;
	 dest[d++]='0';
	 dest[d++]='x';

	 /* length = -1 means assume dest is large enough, ugly */
	 if (destlen==-1) {
          for (s=0;s<srclen;s++,d+=2) {
             dest[d]=src[s]/16 > 9 ? 87 + (src[s]/16) : '0' + (src[s]/16);
             dest[d+1]=src[s]%16 > 9 ? 87 + (src[s]%16) : '0' + (src[s]%16);
          }
	 } else {
          for (s=0;s<srclen && d < destlen-2;s++,d+=2) {
             dest[d]=src[s]/16 > 9 ? 87 + (src[s]/16) : '0' + (src[s]/16);
             dest[d+1]=src[s]%16 > 9 ? 87 + (src[s]%16) : '0' + (src[s]%16);
          }
	}
	dest[d++]='\0';
        return d;
      case SYBIMAGE:
      case SYBBINARY:
	 cplen = srclen > destlen ? destlen : srclen;
         memcpy(dest, src, cplen);
	 return cplen;
      case SYBVARBINARY:
	 cplen = srclen > destlen ? destlen : srclen;
	 varbin = (TDS_VARBINARY *) dest;
	 varbin->len = cplen;
         memcpy(varbin->array, src, cplen);
	 return sizeof(TDS_VARBINARY);
   }
   return TDS_FAIL;
}

static TDS_INT 
tds_convert_char(int srctype,TDS_CHAR *src, TDS_UINT srclen,
	int desttype,TDS_CHAR *dest,TDS_INT destlen)
{
DBANY any;
int   ret;
int   i;

   
   switch(desttype) {
      case SYBCHAR:
      case SYBVARCHAR:
      case SYBNVARCHAR:
      case SYBTEXT:
		any.c = malloc(srclen + 1);
		memset(any.c, '\0', srclen + 1);
		memcpy(any.c, src, srclen);
		break;
         break;
      case SYBBINARY:
      case SYBIMAGE:
         break;
      case SYBINT1:
         any.ti = atoi(src);
         break;
      case SYBINT2:
         any.si = atoi(src);
         break;
      case SYBINT4:
         any.i = atoi(src);
         break;
      case SYBFLT8:
         any.f = atof(src);
         break;
      case SYBREAL:
         any.r = atof(src);
         break;
      case SYBBIT:
      case SYBBITN:
         any.ti = (atoi(src)>0) ? 1 : 0;
         break;
      case SYBMONEY:
         break;
      case SYBMONEY4:
         break;
      case SYBDATETIME:
		if (!string_to_datetime(src, SYBDATETIME, &any))
           return TDS_FAIL;
		break;
      case SYBDATETIME4:
		if (!string_to_datetime(src, SYBDATETIME4, &any))
           return TDS_FAIL;
		break;
      case SYBNUMERIC:
      case SYBDECIMAL:
		any.n.precision = ((TDS_NUMERIC *)dest)->precision;
		any.n.scale = ((TDS_NUMERIC *)dest)->scale;
		string_to_numeric(src, &any);
		break;
/*
		case SYBBOUNDRY:
			break;
		case SYBSENSITIVITY:
			break;
*/
		default:
			return TDS_FAIL;
	}

	ret = tds_convert_any(dest, desttype, destlen, &any);

	switch(desttype) {
		case SYBCHAR:
		case SYBVARCHAR:
		case SYBNVARCHAR:
			free(any.c);
	}
	return(ret);
}

static TDS_INT 
tds_convert_bit(int srctype,TDS_CHAR *src,
	int desttype,TDS_CHAR *dest,TDS_INT destlen)
{
DBANY any;

	switch(desttype) {
		case SYBCHAR:
		case SYBVARCHAR:
			sprintf(tmp_str,"%c",src[0] ? '1' : '0');
			any.c = tmp_str;
			break;
		case SYBTEXT:
			break;
		case SYBBINARY:
		case SYBIMAGE:
			break;
		case SYBINT1:
			any.ti = src[0] ? 1 : 0;
			break;
		case SYBINT2:
			any.si = src[0] ? 1 : 0;
			break;
		case SYBINT4:
			any.i = src[0] ? 1 : 0;
			break;
		case SYBFLT8:
			any.f = src[0] ? 1.0 : 0.0;
			break;
		case SYBREAL:
			any.r = src[0] ? 1.0 : 0.0;
			break;
		case SYBBIT:
		case SYBBITN:
			any.ti = src[0];
			break;
		case SYBMONEY:
			break;
		case SYBMONEY4:
			break;
		case SYBNUMERIC:
		case SYBDECIMAL:
			break;
/*
		case SYBBOUNDRY:
			break;
		case SYBSENSITIVITY:
			break;
*/
	}
	return tds_convert_any(dest, desttype, destlen, &any);
}
static TDS_INT 
tds_convert_int1(int srctype,TDS_CHAR *src,
	int desttype,TDS_CHAR *dest,TDS_INT destlen)
{
TDS_TINYINT buf;
DBANY any;

	memcpy(&buf, src, sizeof(buf));
	switch(desttype) {
		case SYBCHAR:
		case SYBVARCHAR:
			sprintf(tmp_str,"%d",buf);
			any.c = tmp_str;
			break;
		case SYBINT1:
			any.ti = buf;
			break;
		case SYBINT2:
			any.si = buf;
			break;
		case SYBINT4:
			any.i = buf;
			break;
		default:
			return TDS_FAIL;
	}
	return tds_convert_any(dest, desttype, destlen, &any);
}
static TDS_INT 
tds_convert_int2(int srctype,TDS_CHAR *src,
	int desttype,TDS_CHAR *dest,TDS_INT destlen)
{
TDS_SMALLINT buf;
DBANY any;
	
	memcpy(&buf,src,sizeof(buf));
	switch(desttype) {
		case SYBCHAR:
		case SYBVARCHAR:
			sprintf(tmp_str,"%d",buf);
			any.c = tmp_str;
			break;
		case SYBINT1:
			any.ti = buf;
			break;
		case SYBINT2:
			any.si = buf;
			break;
		case SYBINT4:
			any.i = buf;
			break;
	}
	return tds_convert_any(dest, desttype, destlen, &any);
}
static TDS_INT 
tds_convert_int4(int srctype,TDS_CHAR *src,
	int desttype,TDS_CHAR *dest,TDS_INT destlen)
{
TDS_INT buf;
DBANY any;
	
	memcpy(&buf,src,sizeof(buf));
	switch(desttype) {
		case SYBCHAR:
		case SYBVARCHAR:
			sprintf(tmp_str,"%d",buf);
			any.c = tmp_str;
			break;
		case SYBINT1:
			any.ti = buf;
			break;
		case SYBINT2:
			any.si = buf;
			break;
		case SYBINT4:
			any.i = buf;
			break;
		default:
			return TDS_FAIL;
	}
	return tds_convert_any(dest, desttype, destlen, &any);
}
static TDS_INT 
tds_convert_numeric(int srctype,TDS_NUMERIC *src,TDS_INT srclen,
	int desttype,TDS_CHAR *dest,TDS_INT destlen)
{
char tmpstr[MAXPRECISION];
TDS_FLOAT d;

	switch(desttype) {
		case SYBCHAR:
		case SYBVARCHAR:
			tds_numeric_to_string(src,dest);
                        return strlen(dest);
			break;
		case SYBNUMERIC:
		case SYBDECIMAL:
			memcpy(dest, src, sizeof(TDS_NUMERIC));
			return sizeof(TDS_NUMERIC);
			break;
		case SYBFLT8:
			/* FIX ME -- temporary fix */
			tds_numeric_to_string(src,tmpstr);
			d = atof(tmpstr);
			memcpy(dest,&d,sizeof(d));
			break;
		default:
			return TDS_FAIL;
			break;
	}
	return TDS_FAIL;
}
static TDS_INT 
tds_convert_money4(int srctype,TDS_CHAR *src, int srclen,
	int desttype,TDS_CHAR *dest,TDS_INT destlen)
{
TDS_MONEY4 mny;
long dollars, fraction;
TDS_FLOAT tmpf;

	switch(desttype) {
		case SYBCHAR:
		case SYBVARCHAR:
			memcpy(&mny, src, sizeof(mny));
			dollars = mny.mny4 / 10000;
			fraction = mny.mny4 % 10000;
			if (fraction < 0)	{ fraction = -fraction; }
			sprintf(dest,"%ld.%02lu",dollars,fraction/100);
			if (desttype==SYBCHAR) {
				/* ugly hack to emulate sybase behaviour
				** which PHP relies on */
				if (destlen==-1)
					_tds_pad_string(dest,srclen);
				else
					_tds_pad_string(dest,destlen);
			}
			return strlen(dest);
			break;
		case SYBFLT8:
			memcpy(&dollars, src, sizeof(dollars));
			tmpf = ((TDS_FLOAT)dollars) / 10000;
			memcpy(dest, &tmpf, sizeof(tmpf));
			break;
		case SYBMONEY4:
			memcpy(dest, src, sizeof(TDS_MONEY4));
			break;
                default:
                        return TDS_FAIL;
                        break;
	}
	return sizeof(TDS_MONEY4);
}
static TDS_INT 
tds_convert_money(int srctype,TDS_CHAR *src,
	int desttype,TDS_CHAR *dest,TDS_INT destlen)
{
TDS_INT high;
TDS_UINT low;
double dmoney;
char *s;

	switch(desttype) {
		case SYBCHAR:
		case SYBVARCHAR:
			/* begin lkrauss 2001-10-13 - fix return to be strlen() */
			s = tds_money_to_string((TDS_MONEY *)src, dest);
			return strlen(s);
			break;
		case SYBFLT8:
			/* Used memcpy to avoid alignment/bus errors */
			memcpy(&high, src, 4);
			memcpy(&low, src+4, 4);
			dmoney = (TDS_FLOAT)high * 65536.0 * 65536.0 + (TDS_FLOAT)low;
			dmoney = dmoney / 10000;
			memcpy(dest, &dmoney, sizeof(TDS_FLOAT));
			*(TDS_FLOAT *)dest = dmoney; /* in case memcpy didn't work? */
			return sizeof(TDS_FLOAT); /* was returning FAIL below (mlilback, 11/7/01) */
			break;
		case SYBMONEY:
			memcpy(dest, src, sizeof(TDS_MONEY));
			break;
	       default:
			return TDS_FAIL;
			break;
	}
	return TDS_FAIL;
}
static TDS_INT 
tds_convert_datetime(TDSLOCINFO *locale, int srctype,TDS_CHAR *src,
	int desttype,TDS_CHAR *dest,TDS_INT destlen)
{

DBANY any;

unsigned int dt_days, dt_time;
int  dim[12]   = { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };

int dty;

char whole_date_string[30];
struct tds_tm when;

TDS_INT ret;

	switch(desttype) {
		case SYBCHAR:
		case SYBVARCHAR:
		case SYBNVARCHAR:
		case SYBTEXT:
			if (!src) {
				any.c = malloc(1);
				*(any.c) = '\0';
			} else {
				/*
				 * Set up an extended struct tm, and call tds_strftime()
				 * to format the datetime as a string.
				 */
				memset( &when, 0, sizeof(when) );

				memcpy(&dt_days, src, 4);
				memcpy(&dt_time, src + 4, 4);

				/* its a date before 1900 */
				if (dt_days > 2958463) { 	/* what's 2958463? */
					dt_days = 0xffffffff  - dt_days; 
					/* year */
					when.tm.tm_year = -1;
					dty = days_this_year(when.tm.tm_year);
					while ( dt_days >= dty ) {
						when.tm.tm_year--;
						dt_days -= dty;
						dty = days_this_year(when.tm.tm_year);
					}
					dim[1] = (dty == 366)? 29 : 28;

					/* month, day */
					when.tm.tm_mon = 11;
					while (dt_days > dim[when.tm.tm_mon] ) {
						dt_days -= dim[when.tm.tm_mon];
						when.tm.tm_mon--;
					}

					when.tm.tm_mday = dim[when.tm.tm_mon] - dt_days;
				} else {
					dt_days++;
					/* year */
					when.tm.tm_year = 0;
					dty = days_this_year(when.tm.tm_year);
					while ( dt_days > dty ) {
						when.tm.tm_year++;
						dt_days -= dty;
						dty = days_this_year(when.tm.tm_year);
					}

					dim[1] = (dty == 366)? 29 : 28;

					/* month, day */
					when.tm.tm_mon = 0;
					while (dt_days > dim[when.tm.tm_mon] ) {
						dt_days -= dim[when.tm.tm_mon];
						when.tm.tm_mon++;
					}
					when.tm.tm_mday = dt_days;
				}
				when.tm.tm_sec = dt_time / 300;
				when.milliseconds = ((dt_time - (when.tm.tm_sec * 300)) * 1000) / 300 ;

				/* hours, minutes, seconds */
				when.tm.tm_hour = when.tm.tm_sec / 3600; 
				when.tm.tm_min = (when.tm.tm_sec % 3600) / 60; 
				when.tm.tm_sec =  when.tm.tm_sec %   60; 

				tds_strftime( whole_date_string, sizeof(whole_date_string), locale->date_fmt, &when );
				any.c = malloc (strlen(whole_date_string) + 1);
				strcpy(any.c , whole_date_string);
			}
			break;
		case SYBDATETIME:
			memcpy(&dt_days, src, 4);
			memcpy(&dt_time, src + 4, 4);
			any.dt.dtdays = dt_days;
			any.dt.dttime = dt_time;
			break;
		case SYBDATETIME4:
			memcpy(&dt_days, src, 4);
			memcpy(&dt_time, src + 4, 4);
			any.dt4.days    = dt_days;
			any.dt4.minutes = (dt_time / 300) / 60;
			break;
		default:
			return TDS_FAIL;
			break;
	}

	ret = tds_convert_any(dest, desttype, destlen, &any);

	switch(desttype) {
		case SYBCHAR:
		case SYBVARCHAR:
		case SYBNVARCHAR:
		case SYBTEXT:
			free(any.c);
	}
	return(ret);
}

int days_this_year (int years)
{
int year;

   year = 1900 + years;
   if ((year % 4 == 0 && year % 100 != 0) || year % 400 == 0)
      return 366;
   else
      return 365;
}

static TDS_INT 
tds_convert_datetime4(TDSLOCINFO *locale, int srctype, TDS_CHAR *src,
	int desttype, TDS_CHAR *dest,TDS_INT destlen)
{

TDS_USMALLINT dt_days, dt_mins;
int  dim[12]   = { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };

int dty = 365;

char whole_date_string[30];
struct tds_tm when;

TDS_INT ret;

DBANY any;

	switch(desttype) {
		case SYBCHAR:
		case SYBVARCHAR:
		case SYBNVARCHAR:
		case SYBTEXT:
			if (!src) {
				any.c = malloc(1);
				*(any.c) = '\0';
			} else {
				/*
				 * Set up an extended struct tm, and call tds_strftime()
				 * to format the datetime as a string.
				 */
				memset( &when, 0, sizeof(when) );

				memcpy(&dt_days, src, 2);
				memcpy(&dt_mins, src + 2, 2);

				dt_days++;
				
				/* year */
				while ( dt_days > dty ) {
					when.tm.tm_year++;
					dt_days -= dty;
					dty = days_this_year(when.tm.tm_year);
				}

				dim[1] = (dty == 366)? 29 : 28;		/* February */

				/* month, day */
				while (dt_days > dim[when.tm.tm_mon] ) {
					dt_days -= dim[when.tm.tm_mon];
					when.tm.tm_mon++;
				}
				when.tm.tm_mday = dt_days;

				/* hours, minutes */
				when.tm.tm_hour = dt_mins / 60; 
				when.tm.tm_min =  dt_mins % 60; 

				/* no seconds, milliseconds for smalldatetime */

				tds_strftime( whole_date_string, sizeof(whole_date_string), locale->date_fmt, &when );

				any.c = malloc (strlen(whole_date_string) + 1);
				strcpy(any.c , whole_date_string);
			}
			break;
		case SYBDATETIME:
			memcpy(&dt_days, src, 2);
			memcpy(&dt_mins, src + 2, 2);
			any.dt.dtdays = dt_days;
			any.dt.dttime = (dt_mins * 60) * 300;
			break;
		case SYBDATETIME4:
			memcpy(&dt_days, src, 2);
			memcpy(&dt_mins, src + 2, 2);
			any.dt4.days    = dt_days;
			any.dt4.minutes = dt_mins;
			break;
		default:
			return TDS_FAIL;
			break;
	}

	ret = tds_convert_any(dest, desttype, destlen, &any);

	switch(desttype) {
		case SYBCHAR:
		case SYBVARCHAR:
		case SYBNVARCHAR:
		case SYBTEXT:
			free(any.c);
		}

    return ret;

}

static TDS_INT 
tds_convert_real(int srctype, TDS_CHAR *src,
	int desttype, TDS_CHAR *dest,TDS_INT destlen)
{
TDS_REAL the_value;

   memcpy(&the_value, src, 4);
   switch(desttype) {
      case SYBCHAR:
      case SYBVARCHAR:
      /* A real type has a 7 digit precision plus a two digit exponent
       *   "-d.dddddde+dd"
       * so it should have >=14 bytes allocated (one for the terminating
       * NULL), right?
       */
         if (destlen >= 0 && destlen < 14)  return TDS_FAIL;
         sprintf(dest,"%.7g", the_value);
         return strlen(dest);
      case SYBFLT8:
         *((TDS_FLOAT *)dest) = the_value;
	 return sizeof(TDS_FLOAT);
      case SYBREAL:
	 memcpy(dest,src,sizeof(TDS_REAL));
         return sizeof(TDS_REAL);
   }
   return TDS_FAIL;
}

static TDS_INT 
tds_convert_flt8(int srctype, TDS_CHAR *src,
	int desttype, TDS_CHAR *dest,TDS_INT destlen)
{
TDS_FLOAT the_value;

   memcpy(&the_value, src, 8);
   switch(desttype) {
      case SYBCHAR:
      case SYBVARCHAR:
      /* A flt8 type has a 15 digit precision plus a three digit exponent
       *   "-d.dddddddddddddde+ddd"
       * so it should have >=23 bytes allocated (one for the terminating
       * NULL), right?
       */
         if (destlen >= 0 && destlen < 23)  return TDS_FAIL;
         sprintf(dest,"%.15g", the_value);
         return strlen(dest);
      case SYBREAL:
         *((TDS_REAL *)dest) = the_value;
	 return sizeof(TDS_REAL);
      case SYBFLT8:
	 memcpy(dest,src,sizeof(TDS_FLOAT));
	 return sizeof(TDS_FLOAT);
   }
   return TDS_FAIL;
}
tds_convert_unique(int srctype,TDS_CHAR *src, TDS_INT srclen,
	int desttype,TDS_CHAR *dest,TDS_INT destlen)
{

TDS_UCHAR buf[16];
DBANY any;

    memcpy(buf, src, 16);

	switch(desttype) {
		case SYBCHAR:
		case SYBVARCHAR:
            any.c = malloc(37);
			sprintf(any.c,"%02X%02X%02X%02X-%02X%02X-%02X%02X-%02X%02X-%02X%02X%02X%02X%02X%02X",
                    buf[0], buf[1], buf[2], buf[3],
                    buf[4], buf[5],
                    buf[6], buf[7],
                    buf[8], buf[9],
                    buf[10], buf[11], buf[12], buf[13], buf[14], buf[15]);
			break;
		case SYBUNIQUE:
            memcpy (&(any.u), src, sizeof(TDS_UNIQUE));
			break;
		default:
			return TDS_FAIL;
	}
	return tds_convert_any(dest, desttype, destlen, &any);
}

TDS_INT 
tds_convert_any(unsigned char *dest, TDS_INT dtype, TDS_INT dlen, DBANY *any)
{
int i;
int ret = TDS_FAIL;

	switch(dtype) {
		case SYBCHAR:
		case SYBVARCHAR:
		case SYBTEXT:
			tdsdump_log(TDS_DBG_INFO1, "%L converting string dlen = %d dtype = %d string = %s\n",dlen,dtype,any->c);
            if (dlen > 0) {
               if (strlen(any->c) > dlen ) {
                  strncpy(dest, any->c , dlen);
                  ret = dlen;
               }
               else {
                  ret = strlen(any->c);
                  strncpy(dest, any->c, ret);
                  for ( i = ret; i < dlen; i++)
                     dest[i] = ' ';
               }
            }
            if (dlen == -1 ) {
               strcpy(dest, any->c);
                  for (i=strlen(dest)-1;dest[i]==' ';i--)
                      dest[i]='\0';
               ret = strlen(dest);
            }
            if (dlen == -2 ) {
               strcpy(dest, any->c);
               ret = strlen(dest);
            }
			break;
		case SYBBINARY:
		case SYBIMAGE:
			break;
		case SYBINT1:
			memcpy(dest,&(any->ti),1);
			return 1;
			break;
		case SYBINT2:
			memcpy(dest,&(any->si),2);
			return 2;
			break;
		case SYBINT4:
			memcpy(dest,&(any->i),4);
			return 4;
			break;
		case SYBFLT8:
			memcpy(dest,&(any->f),8);
			return 8;	
			break;
		case SYBREAL:
			memcpy(dest,&(any->r),4);
			return 4;	
			break;
		case SYBBIT:
		case SYBBITN:
			memcpy(dest,&(any->ti),1);
			return 1;
			break;
		case SYBMONEY:
			break;
		case SYBMONEY4:
			break;
		case SYBDATETIME:
			memcpy(dest,&(any->dt),sizeof(TDS_DATETIME));
			ret = sizeof(TDS_DATETIME);
			break;
		case SYBDATETIME4:
			memcpy(dest,&(any->dt4),sizeof(TDS_DATETIME4));
			ret = sizeof(TDS_DATETIME4);
			break;
		case SYBNUMERIC:
		case SYBDECIMAL:
			memcpy(dest,&(any->n), sizeof(TDS_NUMERIC));
			ret = sizeof(TDS_NUMERIC);
			break;
/*
		case SYBBOUNDRY:
			break;
		case SYBSENSITIVITY:
			break;
*/
		case SYBUNIQUE:
			memcpy(dest,&(any->u), sizeof(TDS_UNIQUE));
			ret = sizeof(TDS_UNIQUE);
			break;
	}
	tdsdump_log(TDS_DBG_INFO1, "%L tds_convert_any returning %d\n", ret);
	return (ret);
}

TDS_INT 
tds_convert(TDSLOCINFO *locale, int srctype, TDS_CHAR *src, TDS_UINT srclen,
		int desttype, TDS_CHAR *dest, TDS_UINT destlen)
{
TDS_VARBINARY *varbin;

	switch(srctype) {
		case SYBCHAR:
		case SYBVARCHAR:
		case SYBNVARCHAR:
			return tds_convert_char(srctype,src, srclen,
				desttype,dest,destlen);
			break;
		case SYBMONEY4:
			return tds_convert_money4(srctype,src,srclen,
				desttype,dest,destlen);
			break;
		case SYBMONEY:
			return tds_convert_money(srctype,src,
				desttype,dest,destlen);
			break;
		case SYBNUMERIC:
		case SYBDECIMAL:
			return tds_convert_numeric(srctype,(TDS_NUMERIC *) src,srclen,
				desttype,dest,destlen);
			break;
		case SYBBIT:
		case SYBBITN:
			return tds_convert_bit(srctype,src,
				desttype,dest,destlen);
			break;
		case SYBINT1:
			return tds_convert_int1(srctype,src,
				desttype,dest,destlen);
			break;
		case SYBINT2:
			return tds_convert_int2(srctype,src,
				desttype,dest,destlen);
			break;
		case SYBINT4:
			return tds_convert_int4(srctype,src,
				desttype,dest,destlen);
			break;
		case SYBREAL:
			return tds_convert_real(srctype,src,
				desttype,dest,destlen);
			break;
		case SYBFLT8:
			return tds_convert_flt8(srctype,src,
				desttype,dest,destlen);
			break;
		case SYBDATETIME:
			return tds_convert_datetime(locale, srctype,src,
				desttype,dest,destlen);
			break;
		case SYBDATETIME4:
			return tds_convert_datetime4(locale, srctype,src,
				desttype,dest,destlen);
			break;
		case SYBVARBINARY:
			varbin = (TDS_VARBINARY *)src;
			return tds_convert_binary(srctype, (TDS_CHAR *)varbin->array,
				varbin->len,desttype, (TDS_CHAR *)dest,destlen);
			break;
		case SYBIMAGE:
		case SYBBINARY:
			return tds_convert_binary(srctype, (TDS_CHAR *)src,srclen,
				desttype, (TDS_CHAR *)dest,destlen);
			break;
		case SYBTEXT:
			return tds_convert_text(srctype,src,srclen,
				desttype,dest,destlen);
			break;
		case SYBNTEXT:
			return tds_convert_ntext(srctype,src,srclen,
				desttype,dest,destlen);
			break;
		case SYBUNIQUE:
			return tds_convert_unique(srctype,src,srclen,
				desttype,dest,destlen);
			break;
		default:
			fprintf(stderr,"Attempting to convert unknown source type %d\n",srctype);

	}
	return TDS_FAIL;
}
static int string_to_datetime(char *instr, int desttype,  DBANY *any)
{
enum states { GOING_IN_BLIND,
              PUT_NUMERIC_IN_CONTEXT,
              DOING_ALPHABETIC_DATE,
              STRING_GARBLED };

char *in;
char *tok;
char last_token[32];
int   monthdone = 0;
int   yeardone  = 0;
int   mdaydone  = 0;
int   timedone  = 0;
int   ampmdone  = 0;

struct tds_time mytime;
struct tds_time *t;

unsigned int dt_days, dt_time;
int          dim[12]   = { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };
int          dty, i;
float        conv_ms;

int current_state;

	memset(&mytime, '\0', sizeof(struct tds_time));
	t = &mytime;

	in = (char *)malloc(strlen(instr));
	strcpy (in , instr );

	tok = strtok (in, " ,");

	current_state = GOING_IN_BLIND;

	while (tok != (char *) NULL) {
		switch (current_state) {
			case GOING_IN_BLIND :
				/* If we have no idea of current context, then if we have */
				/* encountered a purely alphabetic string, it MUST be an  */
				/* alphabetic month name or prefix...                     */

				if (is_alphabetic(tok)) {
					if (is_monthname(tok)) {
						store_monthname(tok, t);
						monthdone++;
						current_state = DOING_ALPHABETIC_DATE;
					} else {
						current_state = STRING_GARBLED;
					}
				}

              /* ...whereas if it is numeric, it could be a number of   */
              /* things...                                              */

              else if (is_numeric(tok)) {
                      switch(strlen(tok)) {
                         /* in this context a 4 character numeric can   */
                         /* ONLY be the year part of an alphabetic date */

                         case 4:
                            store_year(atoi(tok), t);
                            yeardone++;
                            current_state = DOING_ALPHABETIC_DATE;
                            break;

                         /* whereas these could be the hour part of a   */
                         /* time specification ( 4 PM ) or the leading  */
                         /* day part of an alphabetic date ( 15 Jan )   */

                         case 2:
                         case 1:
                            strcpy(last_token, tok);
                            current_state = PUT_NUMERIC_IN_CONTEXT;
                            break;

                         /* this must be a [YY]YYMMDD date             */

                         case 6:
                         case 8:
                            if (store_yymmdd_date(tok, t))
                               current_state = GOING_IN_BLIND;
                            else
                               current_state = STRING_GARBLED;
                            break;

                         /* anything else is nonsense...               */

                         default:
                            current_state = STRING_GARBLED;
                            break;
                      }
                   }

                   /* it could be [M]M/[D]D/[YY]YY format              */

                   else if (is_numeric_dateformat(tok)) {
                            store_numeric_date(tok, t);
                            current_state = GOING_IN_BLIND;
                   } else if (is_timeformat(tok)) {
                                store_time(tok, t);
                                current_state = GOING_IN_BLIND;
                   } else {
                                 current_state = STRING_GARBLED;
						 }

              break;   /* end of GOING_IN_BLIND */                               

           case DOING_ALPHABETIC_DATE:

              if (is_alphabetic(tok)) {
                 if (!monthdone && is_monthname(tok)) {
                     store_monthname(tok, t);
                     monthdone++;
                     if (monthdone && yeardone && mdaydone )
                        current_state = GOING_IN_BLIND;
                     else
                        current_state = DOING_ALPHABETIC_DATE;
                 } else {
                     current_state = STRING_GARBLED;
					  }
              } else if (is_numeric(tok)) {
                      if (mdaydone && yeardone)
                          current_state = STRING_GARBLED;
                      else switch(strlen(tok)) {
                              case 4:
                                 store_year(atoi(tok), t);
                                 yeardone++;
                                 if (monthdone && yeardone && mdaydone )
                                    current_state = GOING_IN_BLIND;
                                 else
                                    current_state = DOING_ALPHABETIC_DATE;
                                 break;

                              case 2:
                              case 1:
                                 if (!mdaydone) {
                                    store_mday(tok, t);

                                    mdaydone++;
                                    if (monthdone && yeardone && mdaydone )
                                       current_state = GOING_IN_BLIND;
                                    else
                                       current_state = DOING_ALPHABETIC_DATE;
                                 } else {
                                    store_year(atoi(tok), t);
                                    yeardone++;
                                    if (monthdone && yeardone && mdaydone )
                                       current_state = GOING_IN_BLIND;
                                    else
                                       current_state = DOING_ALPHABETIC_DATE;
                                 }
                                 break;

                              default:
                                 current_state = STRING_GARBLED;
                           }
                   } else {
                     current_state = STRING_GARBLED;
						 }

              break;   /* end of DOING_ALPHABETIC_DATE */                        

           case PUT_NUMERIC_IN_CONTEXT:

              if (is_alphabetic(tok)) {
                 if (is_monthname(tok)) {
                     store_mday(last_token, t);
                     mdaydone++;
                     store_monthname(tok, t);
                     monthdone++;
                     if (monthdone && yeardone && mdaydone )
                        current_state = GOING_IN_BLIND;
                     else
                        current_state = DOING_ALPHABETIC_DATE;
                 } else if (is_ampm(tok)) {
                     store_hour(last_token, tok, t);
                     current_state = GOING_IN_BLIND;
                 } else {
                         current_state = STRING_GARBLED;
					  }
              } else if (is_numeric(tok)) {
                      switch(strlen(tok)) {
                         case 4:
                         case 2:
                           store_mday(last_token, t);
                           mdaydone++;
                           store_year(atoi(tok), t);
                           yeardone++;

                           if (monthdone && yeardone && mdaydone )
                              current_state = GOING_IN_BLIND;
                           else
                              current_state = DOING_ALPHABETIC_DATE;
                           break;

                         default:
                           current_state = STRING_GARBLED;
                       }
                } else {
                   current_state = STRING_GARBLED;
					 }

              break;   /* end of PUT_NUMERIC_IN_CONTEXT */                       

           case STRING_GARBLED:
              return (0);
        }

        tok = strtok((char *)NULL, " ,");
    }

    /* 1900 or after */ 
    if (t->tm_year >= 0) {
       dt_days = 0;
       for (i = 0; i < t->tm_year ; i++) {
           dty = days_this_year(i);
           dt_days += dty;
       }

       dty = days_this_year(i);
       if (dty == 366 )
           dim[1] = 29;
       else
           dim[1] = 28;
       for (i = 0; i < t->tm_mon ; i++) {
           dt_days += dim[i];
       }

       dt_days += (t->tm_mday - 1);

    } else {
 	   dt_days = 0xffffffff;
       /* dt_days = 4294967295U;  0xffffffff */
       for (i = -1; i > t->tm_year ; i--) {
           dty = days_this_year(i);
           dt_days -= dty;
       }
       dty = days_this_year(i);
       if (dty == 366 )
           dim[1] = 29;
       else
           dim[1] = 28;

       for (i = 11; i > t->tm_mon ; i--) {
           dt_days -= dim[i];
       }

       dt_days -= dim[i] - t->tm_mday;

    }

    if ( desttype == SYBDATETIME ) {
       any->dt.dtdays = dt_days;

       dt_time = 0;

       for (i = 0; i < t->tm_hour ; i++)
           dt_time += 3600;

       for (i = 0; i < t->tm_min ; i++)
           dt_time += 60;

       dt_time += t->tm_sec;

       any->dt.dttime = dt_time * 300;

       conv_ms = ((float)t->tm_ms / 1000) * 300;
       any->dt.dttime += conv_ms;
    } else {
       /* SYBDATETIME4 */ 
       any->dt4.days = dt_days;

       dt_time = 0;

       for (i = 0; i < t->tm_hour ; i++)
           dt_time += 60;

       for (i = 0; i < t->tm_min ; i++)
           dt_time += 1;

        any->dt4.minutes = dt_time;

    }

    free(in);
    return (1);
}

static void 
string_to_numeric(char *instr, DBANY *any)
{

char  mynumber[39];
/* unsigned char  mynumeric[16]; */ 

char *ptr;
char c = '\0';

unsigned char masks[8] = { 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01 };

short int  binarylist[128];
short int  carry_on = 1;
short int  i = 0;
short int  j = 0;
short int  x = 0;
short int bits, bytes, places, point_found, sign;

struct diglist *topptr = (struct diglist *)NULL;
struct diglist *curptr = (struct diglist *)NULL;
struct diglist *freeptr = (struct diglist *)NULL;

  sign        = 0;
  point_found = 0;
  places      = 0;


  if (any->n.precision == 0)
     any->n.precision = 18; 


  for (ptr = instr; *ptr == ' '; ptr++);  /* skip leading blanks */

  if (*ptr == '-' || *ptr == '+')         /* deal with a leading sign */
  {
     if (*ptr == '-')
        sign = 1;
     ptr++;
  }

  for(; *ptr; ptr++)                      /* deal with the rest */
  {
     if (isdigit(*ptr))                   /* its a number */
     {  
        mynumber[i++] = *ptr;
        if (point_found)                  /* if we passed a decimal point */
           places++;                      /* count digits after that point  */
     }
     else if (*ptr == '.')                /* found a decimal point */
          {
             if (point_found)             /* already had one. lose the rest */
                break;
             if (any->n.scale == 0)       /* no scale...lose the rest  */
                break;
             point_found = 1;
          }
          else                            /* first invalid character */
             break;                       /* lose the rest.          */
  }


  if (any->n.scale > 0)                   /* scale specified, pad out */
  {                                       /* number with zeroes to the */
                                          /* scale...                  */

     for (j = 0 ; j < (any->n.scale - places) ; j++ )
         mynumber[i++] = '0';

  }

  mynumber[i] = '\0';


  if (strlen(mynumber) > any->n.precision )
     strcpy(mynumber, &mynumber[strlen(mynumber) - any->n.precision]);
 
  for (ptr = mynumber; *ptr; ptr++)
  {
      if (topptr == (struct diglist *)NULL)
      {
          topptr = (struct diglist *)malloc(sizeof(struct diglist));
          curptr = topptr;
          curptr->nextptr = NULL;
          curptr->dig = *ptr - 48;
          curptr->carried = 0;
      }
      else
      {
          curptr->nextptr = (struct diglist *)malloc(sizeof(struct diglist));
          curptr = curptr->nextptr;
          curptr->nextptr = NULL;
          curptr->dig = *ptr - 48;
          curptr->carried = 0;
      }
  }

  memset(&binarylist[0], '\0',  sizeof(short int) * 128);
  i = 127;

  while (carry_on)
  {
     carry_on = 0;
     for (curptr = topptr ; curptr != NULL; curptr = curptr->nextptr)
     {
         if (curptr->dig > 0)
            carry_on = 1;
     
         if (curptr->nextptr != NULL )
         {
            curptr->nextptr->carried = ( ( curptr->carried * 10 ) + curptr->dig ) % 2;
            curptr->dig =  ( ( curptr->carried * 10 ) + curptr->dig ) / 2;
         }
         else
         {
            if (carry_on)
            {
               binarylist[i--] = ( ( curptr->carried * 10 ) + curptr->dig ) % 2;
            }
            curptr->dig = ( ( curptr->carried * 10 ) + curptr->dig ) / 2;
         }
     }

  }

  memset(any->n.array, '\0', 17);
  bits  = 0;
  bytes = 1;

  any->n.array[0] =  sign;

  x = g__numeric_bytes_per_prec[any->n.precision] - 1;

  for (i = 128 - (x * 8); i < 128 ; i++)
  {
     if (binarylist[i])
     {
        c = c | masks[bits]; 
     }
     bits++;
     if (bits == 8)
     {
        any->n.array[bytes] = c;

        bytes++;
        bits = 0;
        c    = '\0';
     }
  }

  curptr = topptr;
  while (curptr != NULL)
  {
      freeptr = curptr;
      curptr  = curptr->nextptr;
      free(freeptr);
  }
}

static int _tds_pad_string(char *dest, int destlen)
{
int i=0;

	if (destlen>strlen(dest)) {
		for (i=strlen(dest)+1;i<destlen;i++)
			dest[i]=' ';
		dest[i]='\0';
	}
	return i;
}
 

static int is_numeric_dateformat(char *t)
{
char *instr ;
int   ret   = 1;
int   slashes  = 0;
int   hyphens  = 0;
int   periods  = 0;
int   digits   = 0;

    for (instr = t; *instr; instr++ )
    {
        if (!isdigit(*instr) && *instr != '/' && *instr != '-' && *instr != '.' )
        {
            ret = 0;
            break;
        }
        if (*instr == '/' ) slashes++;
        else if (*instr == '-' ) hyphens++;
             else if (*instr == '.' ) periods++;
                  else digits++;
       
    }
    if (hyphens + slashes + periods != 2)
       ret = 0;
    if (hyphens == 1 || slashes == 1 || periods == 1)
       ret = 0;

    if (digits < 4 || digits > 8)
       ret = 0;

    return(ret);

}

static int is_monthname(char *datestr)
{

int ret = 0;

    if (strlen(datestr) == 3)
    {
       if (strcasecmp(datestr,"jan") == 0) ret = 1;
       else if (strcasecmp(datestr,"feb") == 0) ret = 1;
       else if (strcasecmp(datestr,"mar") == 0) ret = 1;
       else if (strcasecmp(datestr,"apr") == 0) ret = 1;
       else if (strcasecmp(datestr,"may") == 0) ret = 1;
       else if (strcasecmp(datestr,"jun") == 0) ret = 1;
       else if (strcasecmp(datestr,"jul") == 0) ret = 1;
       else if (strcasecmp(datestr,"aug") == 0) ret = 1;
       else if (strcasecmp(datestr,"sep") == 0) ret = 1;
       else if (strcasecmp(datestr,"oct") == 0) ret = 1;
       else if (strcasecmp(datestr,"nov") == 0) ret = 1;
       else if (strcasecmp(datestr,"dec") == 0) ret = 1;
       else ret = 0;
    }
    else
    {
       if (strcasecmp(datestr,"january") == 0) ret = 1;
       else if (strcasecmp(datestr,"february") == 0) ret = 1;
       else if (strcasecmp(datestr,"march") == 0) ret = 1;
       else if (strcasecmp(datestr,"april") == 0) ret = 1;
       else if (strcasecmp(datestr,"june") == 0) ret = 1;
       else if (strcasecmp(datestr,"july") == 0) ret = 1;
       else if (strcasecmp(datestr,"august") == 0) ret = 1;
       else if (strcasecmp(datestr,"september") == 0) ret = 1;
       else if (strcasecmp(datestr,"october") == 0) ret = 1;
       else if (strcasecmp(datestr,"november") == 0) ret = 1;
       else if (strcasecmp(datestr,"december") == 0) ret = 1;
       else ret = 0;

    }
    return(ret);

}
static int is_ampm(char *datestr)
{

int ret = 0;

    if (strcasecmp(datestr,"am") == 0) ret = 1;
    else if (strcasecmp(datestr,"pm") == 0) ret = 1;
    else ret = 0;

    return(ret);

}

static int is_alphabetic(char *datestr)
{
char *s;
int  ret = 1;
    for (s = datestr; *s; s++) {
        if (!isalpha(*s))
           ret = 0; 
    }
    return(ret);
}

static int is_numeric(char *datestr)
{
char *s;
int  ret = 1;
    for (s = datestr; *s; s++) {
        if (!isdigit(*s))
           ret = 0; 
    }
    return(ret);
}

static int is_timeformat(char *datestr)
{
char *s;
int  ret = 1;
    for (s = datestr; *s; s++) 
    {
        if (!isdigit(*s) && *s != ':' && *s != '.' )
          break;
    }
    if ( *s )
    {
       if (strcasecmp(s, "am" ) != 0 && strcasecmp(s, "pm" ) != 0 )
          ret = 0; 
    }
    

    return(ret);
}

static int store_year(int year , struct tds_time *t)
{

    if ( year <= 0 )
       return 0; 

    if ( year < 100 )
    {
       if (year > 49)
          t->tm_year = year;
       else
          t->tm_year = 100 + year ;
       return (1);
    }

    if ( year < 1753 )
       return (0);

    if ( year <= 9999 )
    {
       t->tm_year = year - 1900;
       return (1);
    }

    return (0);

}
static int store_mday(char *datestr , struct tds_time *t)
{
int  mday = 0;

    mday = atoi(datestr);

    if ( mday > 0 && mday < 32 )
    {
       t->tm_mday = mday;
       return (1);
    }
    else
       return 0; 
}

static int store_numeric_date(char *datestr , struct tds_time *t)
{
enum {TDS_MONTH, 
      TDS_DAY, 
      TDS_YEAR};

int  state = TDS_MONTH;
char last_char = 0; 
char *s;
int  month = 0, year = 0, mday = 0;

    for (s = datestr; *s; s++) {
        if (! isdigit(*s) && isdigit(last_char)) {
            state++;
        } else switch(state) {
            case TDS_MONTH:
                month = (month * 10) + (*s - '0');
                break;
            case TDS_DAY:
                mday = (mday * 10) + (*s - '0');
                break;
            case TDS_YEAR:
                year = (year * 10) + (*s - '0');
                break;
        }
        last_char = *s;
    }

    if ( month > 0 && month < 13 )
       t->tm_mon = month - 1;
    else
       return 0; 
    if ( mday > 0 && mday < 32 )
       t->tm_mday = mday;
    else
       return 0; 

    return store_year(year, t);

}

static int store_monthname(char *datestr , struct tds_time *t)
{

int ret = 0;

    if (strlen(datestr) == 3)
    {
       if (strcasecmp(datestr,"jan") == 0) t->tm_mon = 0;
       else if (strcasecmp(datestr,"feb") == 0) t->tm_mon = 1;
       else if (strcasecmp(datestr,"mar") == 0) t->tm_mon = 2;
       else if (strcasecmp(datestr,"apr") == 0) t->tm_mon = 3;
       else if (strcasecmp(datestr,"may") == 0) t->tm_mon = 4;
       else if (strcasecmp(datestr,"jun") == 0) t->tm_mon = 5;
       else if (strcasecmp(datestr,"jul") == 0) t->tm_mon = 6;
       else if (strcasecmp(datestr,"aug") == 0) t->tm_mon = 7;
       else if (strcasecmp(datestr,"sep") == 0) t->tm_mon = 8;
       else if (strcasecmp(datestr,"oct") == 0) t->tm_mon = 9;
       else if (strcasecmp(datestr,"nov") == 0) t->tm_mon = 10;
       else if (strcasecmp(datestr,"dec") == 0) t->tm_mon = 11;
       else ret = 0;
    }
    else
    {
       if (strcasecmp(datestr,"january") == 0) t->tm_mon = 0;
       else if (strcasecmp(datestr,"february") == 0) t->tm_mon = 1;
       else if (strcasecmp(datestr,"march") == 0) t->tm_mon = 2;
       else if (strcasecmp(datestr,"april") == 0) t->tm_mon = 3;
       else if (strcasecmp(datestr,"june") == 0) t->tm_mon = 5;
       else if (strcasecmp(datestr,"july") == 0) t->tm_mon = 6;
       else if (strcasecmp(datestr,"august") == 0) t->tm_mon = 7;
       else if (strcasecmp(datestr,"september") == 0) t->tm_mon = 8;
       else if (strcasecmp(datestr,"october") == 0) t->tm_mon = 9;
       else if (strcasecmp(datestr,"november") == 0) t->tm_mon = 10;
       else if (strcasecmp(datestr,"december") == 0) t->tm_mon = 11;
       else ret = 0;

    }
    return(ret);

}
static int store_yymmdd_date(char *datestr , struct tds_time *t)
{
int  month = 0, year = 0, mday = 0;

int wholedate;

    wholedate = atoi(datestr);

    year  = wholedate / 10000 ;
    month = ( wholedate - (year * 10000) ) / 100 ; 
    mday  = ( wholedate - (year * 10000) - (month * 100) );

    if ( month > 0 && month < 13 )
       t->tm_mon = month - 1;
    else
       return 0; 
    if ( mday > 0 && mday < 32 )
       t->tm_mday = mday;
    else
       return 0; 

    return (store_year(year, t));

}

static int store_time(char *datestr , struct tds_time *t)
{
enum {TDS_HOURS, 
      TDS_MINUTES, 
      TDS_SECONDS,
      TDS_FRACTIONS};

int  state = TDS_HOURS;
char last_sep;
char *s;
int hours = 0, minutes = 0, seconds = 0, millisecs = 0;
int ret = 1;

    for (s = datestr; 
         *s && strchr("apmAPM" , (int) *s) == (char *)NULL; 
         s++) 
    {
        if ( *s == ':' || *s == '.' ) {
            last_sep = *s;
            state++;
        } else switch(state) {
            case TDS_HOURS:
                hours = (hours * 10) + (*s - '0');
                break;
            case TDS_MINUTES:
                minutes = (minutes * 10) + (*s - '0');
                break;
            case TDS_SECONDS:
                seconds = (seconds * 10) + (*s - '0');
                break;
            case TDS_FRACTIONS:
                millisecs = (millisecs * 10) + (*s - '0');
                break;
        }
    }
    if (*s)
    {
       if(strcasecmp(s,"am") == 0)
       {
          if (hours == 12)
              hours = 0;

          t->tm_hour = hours;
       }
       if(strcasecmp(s,"pm") == 0)
       {
          if (hours == 0)
              ret = 0;
          if (hours > 0 && hours < 12)
              t->tm_hour = hours + 12;
          else
              t->tm_hour = hours;
       }
    }
    else
    {
      if (hours >= 0 && hours < 24 )
         t->tm_hour = hours;
      else
         ret = 0;
    }
    if (minutes >= 0 && minutes < 60)
      t->tm_min = minutes;
    else
      ret = 0;
    if (seconds >= 0 && minutes < 60)
      t->tm_sec = seconds;
    else
      ret = 0;
    if (millisecs)
    {
      if (millisecs >= 0 && millisecs < 1000 )
      {
         if (last_sep == ':')
            t->tm_ms = millisecs;
         else
         {

            if (millisecs < 10)
               t->tm_ms = millisecs * 100;
            else if (millisecs < 100 )
                    t->tm_ms = millisecs * 10;
                 else 
                    t->tm_ms = millisecs;
         }
      }
      else
        ret = 0;
    }


    return (ret);
}

static int store_hour(char *hour , char *ampm , struct tds_time *t)
{
int ret = 1;
int  hours;

    hours = atoi(hour);

    if (hours >= 0 && hours < 24 )
    {
       if(strcasecmp(ampm,"am") == 0)
       {
          if (hours == 12)
              hours = 0;

          t->tm_hour = hours;
       }
       if(strcasecmp(ampm,"pm") == 0)
       {
          if (hours == 0)
              ret = 0;
          if (hours > 0 && hours < 12)
              t->tm_hour = hours + 12;
          else
              t->tm_hour = hours;
       }
    }
    return (ret);
}

TDS_INT tds_get_null_type(int srctype)
{

	switch(srctype) {
		case SYBCHAR:
			return SYBVARCHAR;
			break;
		case SYBINT1:
		case SYBINT2:
		case SYBINT4:
			return SYBINTN;
			break;
		case SYBREAL:
		case SYBFLT8:
			return SYBFLTN;
			break;
		case SYBDATETIME:
		case SYBDATETIME4:
			return SYBDATETIMN;
			break;
		default:
			return srctype;

	}
	return srctype;
}
 
/*
 * format a date string according to an "extended" strftime formatting definition.
 */     
static size_t 
tds_strftime(char *buf, size_t maxsize, const char *format, const struct tds_tm *timeptr)
{
	int length=0;
	char *s, *our_format;
	char millibuf[8];
	
	char *pz = NULL;
	
	our_format = malloc( strlen(format) + 1 );
	if( !our_format ) return 0;
	strcpy( our_format, format );
		
	pz = strstr( our_format, "%z" );
	
	/* 
	 * Look for "%z" in the format string.  If found, replace it with timeptr->milliseconds.
	 * For example, if milliseconds is 124, the format string 
	 * "%b %d %Y %H:%M:%S.%z" would become 
	 * "%b %d %Y %H:%M:%S.124".  
	 */
	 
	/* Skip any escaped cases (%%z) */
	while( pz && *(pz-1) == '%' )
		pz = strstr( ++pz, "%z" );
	
	if( pz && length < maxsize - 1 ) {
		
		sprintf( millibuf, "%03d", timeptr->milliseconds );
		
		/* move everything back one, then overwrite "?%z" with millibuf */
		for( s = our_format + strlen(our_format); s > pz; s-- ) {
			*(s+1) = *s;
		}
		
		strncpy( pz, millibuf, 3 );	/* don't copy the null */
	}
	
	length = strftime( buf, maxsize, our_format, &timeptr->tm );	
	
	free(our_format);
	
	return length;
}
